#include "LZActorValuePercentageFix.h"

namespace lznsAVFix {
	/**
	 * Provides a sneaky way to extend Actor's functionality without recompiling SKSE itself.
	 * The GetActorValueModifier() function **IS** a member of Actor (as determined by reverse engineering),
	 *  but the SKSE team never identified or included this function as of the latest release.
	 * Ideally, the the function should be appended to Actor within skse/GameReferences.h like so:
	 *
	 *     DEFINE_MEMBER_FN(GetActorValueModifier, float, 0x6DE980, UInt32 modType, UInt32 actorValue);
	 *
	 * However, the above would require SKSE to be recompiled, which is not ideal.
	 * Creating a subclass "ActorExtended" and dynamically casting from Actor* to ActorExtended*
	 *  does NOT work because such a cast produces undefined behavior per the C++ specification.
	 * In most cases, attempting the latter will cause a crash.
	 *
	 * The workaround is to first extend Actor as shown below.
	 * All subsequent functions must then assume the extended type, instead of the original type;
	 *  no casting needs to be performed.
	 * This behavior has been verified working both on engine functions and when passing actor instances
	 *  within Papyrus scripts (again, no casting is needed).
	 *
	 * The reason this works is that SKSE never directly manipulates the Actor type; not even in Papyrus.
	 * Instead, raw pointers from memory (see assembly below) are passed as parameters to all SKSE functions,
	 *  and are simply assumed (as determined by reverse engineering) to be Actors.
	 * As long as the extended member function is **truly** a member of Skyrim's underlying actor representation,
	 *  it is correct to assume these raw pointers are the extended Actor class, rather than SKSE's implementation.
	 * In other words, all skse/Actor instances ARE actually of the extended type below.
	 * Direct casts do not work because in essence, you would be asking C++ to cast an object from itself to itself
	 *  (implying a type is actually two different types, which is impossible - hence the undefined behavior).
	 */
	class ActorExtended : public ::Actor {
	public:
		MEMBER_FN_PREFIX(ActorExtended);
		DEFINE_MEMBER_FN(GetActorValueModifier, float, 0x6DE980, UInt32 modType, UInt32 actorValue);
	};

	static const UInt32 Condition_GetAVPercent_Base = 0x5067DF;
	static const UInt32 Condition_GetAVPercent_Return = Condition_GetAVPercent_Base + 0x5;

	static const UInt32 Condition_GetStaminaPercentage_Base = 0x506A0C;
	static const UInt32 Condition_GetStaminaPercentage_Return = Condition_GetStaminaPercentage_Base + 0x5;

	static const UInt32 Condition_GetHealthPercentage_Base = 0x506A8C;
	static const UInt32 Condition_GetHealthPercentage_Return = Condition_GetHealthPercentage_Base + 0x5;

	static const UInt32 Condition_GetAVPercentagePapyrus_Base = 0x8DA706;
	static const UInt32 Condition_GetAVPercentagePapyrus_Return = Condition_GetAVPercentagePapyrus_Base + 0x5;

	/**
	 * Obtains the actor value percent by incorporating the temporary modifier (normally omitted by the game).
	 * The offsets defined above account for all Condition Functions in the Creation Kit, and the associated item for Papyrus.
	 */
	static float GetAVPercent(ActorExtended* actor, UInt32 avIndex) {
		if (actor) {
			const float max = actor->actorValueOwner.GetMaximum(avIndex);
			const float temp = CALL_MEMBER_FN(actor, GetActorValueModifier)(1, avIndex);
			const float sum = max + temp;
			if (sum > 0) {
				return actor->actorValueOwner.GetCurrent(avIndex) / sum;
			}
		}
		return 1.0f;
	}

	/** Overrides the engine behavior for GetActorValuePercent. */
	static void __declspec(naked) hookConditionAVPercent() {
		__asm {
			push eax
			push ecx
			call GetAVPercent
			add esp, 0x10

			jmp Condition_GetAVPercent_Return
		}
	}

	/** Overrides the engine behavior for GetStaminaPercentage. */
	static void __declspec(naked) hookConditionStaminaPercentage() {
		__asm {
			push 0x1A
			push ecx
			call GetAVPercent
			add esp, 0xC

			jmp Condition_GetStaminaPercentage_Return
		}
	}

	/** Overrides the engine behavior for GetStaminaPercentage. */
	static void __declspec(naked) hookConditionHealthPercentage() {
		__asm {
			push 0x18
			push ecx
			call GetAVPercent
			add esp, 0xC

			jmp Condition_GetHealthPercentage_Return
		}
	}

	/** Overrides the engine behavior for the GetActorValuePercentage function in Papyrus. */
	static void __declspec(naked) hookConditionAVPercentPapyrus() {
		__asm {
			push eax
			push ecx
			call GetAVPercent
			add esp, 0x10

			jmp Condition_GetAVPercentagePapyrus_Return
		}
	}

	/** Obtains the actor's temporary AV modifier. */
	float GetAVModifier(StaticFunctionTag* base, ActorExtended* actor, UInt32 modType, BSFixedString actorValue) {
		if (actor) {
			UInt32 av = ActorValueList::ResolveActorValueByName(actorValue.data);
			if (av != 255) {
				return CALL_MEMBER_FN(actor, GetActorValueModifier)(modType, av);
			}
		}
		return 0.0f;
	}

	/** Installs the core mod. */
	void Install() {
		WriteRelCall(Condition_GetAVPercent_Base, (UInt32) hookConditionAVPercent);
		WriteRelCall(Condition_GetStaminaPercentage_Base, (UInt32) hookConditionStaminaPercentage);
		WriteRelCall(Condition_GetHealthPercentage_Base, (UInt32) hookConditionHealthPercentage);
		WriteRelCall(Condition_GetAVPercentagePapyrus_Base, (UInt32) hookConditionAVPercentPapyrus);
	}

	/** Registers all new Papyrus functions into the plugin. */
	bool RegisterFuncs(VMClassRegistry* registry) {
		registry->RegisterFunction(
			new NativeFunction3 <StaticFunctionTag, float, ActorExtended*, UInt32, BSFixedString>("GetAVModifier", "LZAVPF", lznsAVFix::GetAVModifier, registry));

		return true;
	}
}
